/*
Copyright (c) 2019-2023 Digital Energy Cloud Solutions LLC. All Rights Reserved.
Authors:
Petr Krutov, <petr.krutov@digitalenergy.online>
Stanislav Solovev, <spsolovev@digitalenergy.online>
Kasim Baybikov, <kmbaybikov@basistech.ru>
Tim Tkachev, <tvtkachev@basistech.ru>
Sergey Kisil, <svkisil@digitalenergy.online>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/*
Terraform DECORT provider - manage resources provided by DECORT (Digital Energy Cloud
Orchestration Technology) with Terraform by Hashicorp.

Source code: https://repository.basistech.ru/BASIS/terraform-provider-decort

Please see README.md to learn where to place source code so that it
builds seamlessly.

Documentation: https://repository.basistech.ru/BASIS/terraform-provider-decort/wiki
*/

package flipgroup

import (
	"context"
	"fmt"
	"github.com/hashicorp/terraform-plugin-sdk/v2/diag"
	"github.com/hashicorp/terraform-plugin-sdk/v2/helper/schema"
	"repository.basistech.ru/BASIS/decort-golang-sdk/pkg/cloudbroker/flipgroup"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/constants"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/controller"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/dc"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/status"

	log "github.com/sirupsen/logrus"
)

func resourceFlipgroupCreate(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceFlipgroupCreate called with name: %s, accountID %v", d.Get("name").(string), d.Get("account_id").(int))

	c := m.(*controller.ControllerCfg)

	if diags := checkParamsExistence(ctx, d, c); diags != nil {
		return diags
	}

	req := flipgroup.CreateRequest{
		Name:       d.Get("name").(string),
		NetType:    d.Get("net_type").(string),
		ClientType: d.Get("client_type").(string),
		AccountID:  uint64(d.Get("account_id").(int)),
		NetID:      uint64(d.Get("net_id").(int)),
	}

	if IP, ok := d.GetOk("ip"); ok {
		req.IP = IP.(string)
	}
	if description, ok := d.GetOk("desc"); ok {
		req.Description = description.(string)
	}

	resp, err := c.CloudBroker().FLIPGroup().Create(ctx, req)
	if err != nil {
		d.SetId("")
		return diag.FromErr(err)
	}

	d.SetId(fmt.Sprint(resp.ID))
	d.Set("flipgroup_id", resp.ID)

	var warnings dc.Warnings

	if client_ids, ok := d.GetOk("client_ids"); ok {
		casted := client_ids.([]interface{})
		addComputesAfterCreation(ctx, &warnings, c, casted, resp.ID)
	}

	return append(warnings.Get(), resourceFlipgroupRead(ctx, d, m)...)
}

func resourceFlipgroupRead(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceFlipgroupRead: called for flipgroup_id %s, name %s",
		d.Id(), d.Get("name").(string))

	fg, err := utilityFlipgroupCheckPresence(ctx, d, m)
	if err != nil {
		d.SetId("")
		return diag.FromErr(err)
	}

	switch fg.Status {
	case status.Destroyed:
		d.SetId("")
		return diag.Errorf("The flipgroup status is destroyed and cannot be read.")
	}

	flattenFlipgroup(d, fg)

	log.Debugf("resourceFlipgroupRead: after flattenFlipgroup: flipgroup_id %s, name %s",
		d.Id(), d.Get("name").(string))

	return nil
}

func resourceFlipgroupUpdate(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceFlipgroupUpdate called with id: %v", d.Get("flipgroup_id").(int))
	c := m.(*controller.ControllerCfg)

	if diags := checkParamsExistence(ctx, d, c); diags != nil {
		return diags
	}

	fg, err := utilityFlipgroupCheckPresence(ctx, d, m)
	if err != nil {
		d.SetId("")
		return diag.FromErr(err)
	}

	switch fg.Status {
	case status.Destroyed:
		d.SetId("")
		return diag.Errorf("The flipgroup status is destroyed and cannot be updated.")
	}

	var warnings dc.Warnings
	basicUpdate := false
	req := flipgroup.EditRequest{FLIPGroupID: fg.ID}

	if d.HasChange("desc") {
		req.Description = d.Get("desc").(string)
		basicUpdate = true
	}

	if d.HasChange("name") {
		req.Name = d.Get("name").(string)
		basicUpdate = true
	}

	if basicUpdate {
		_, err := c.CloudBroker().FLIPGroup().Edit(ctx, req)
		if err != nil {
			return diag.FromErr(err)
		}
	}

	if d.HasChange("client_ids") {
		handleClientIDsUpdate(ctx, d, c, fg, &warnings)
	}

	return append(warnings.Get(), resourceFlipgroupRead(ctx, d, m)...)
}

func resourceFlipgroupDelete(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceFlipgroupDelete called with id: %v", d.Get("flipgroup_id").(int))
	c := m.(*controller.ControllerCfg)

	fg, err := utilityFlipgroupCheckPresence(ctx, d, m)
	if err != nil {
		d.SetId("")
		return diag.FromErr(err)
	}

	req := flipgroup.DeleteRequest{
		FLIPGroupID: fg.ID,
	}

	// When FLIPGroup().Delete() is executed, flipgroup automatically is removed from the compute it has been attached to, if any.
	// No need to specifically call for FlipGroup().ComputeRemove().
	_, err = c.CloudBroker().FLIPGroup().Delete(ctx, req)
	if err != nil {
		return diag.FromErr(err)
	}

	d.SetId("")

	return nil
}

func ResourceFlipgroup() *schema.Resource {
	return &schema.Resource{
		SchemaVersion: 1,

		Importer: &schema.ResourceImporter{
			StateContext: schema.ImportStatePassthroughContext,
		},

		CreateContext: resourceFlipgroupCreate,
		ReadContext:   resourceFlipgroupRead,
		UpdateContext: resourceFlipgroupUpdate,
		DeleteContext: resourceFlipgroupDelete,

		Timeouts: &schema.ResourceTimeout{
			Create:  &constants.Timeout300s,
			Read:    &constants.Timeout300s,
			Update:  &constants.Timeout300s,
			Delete:  &constants.Timeout300s,
			Default: &constants.Timeout300s,
		},

		Schema: resourceFlipgroupSchemaMake(),
	}
}
