/*
Copyright (c) 2019-2023 Digital Energy Cloud Solutions LLC. All Rights Reserved.
Authors:
Petr Krutov, <petr.krutov@digitalenergy.online>
Stanislav Solovev, <spsolovev@digitalenergy.online>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/*
Terraform DECORT provider - manage resources provided by DECORT (Digital Energy Cloud
Orchestration Technology) with Terraform by Hashicorp.

Source code: https://repository.basistech.ru/BASIS/terraform-provider-decort

Please see README.md to learn where to place source code so that it
builds seamlessly.

Documentation: https://repository.basistech.ru/BASIS/terraform-provider-decort/wiki
*/

package image

import (
	"context"
	"fmt"
	"strconv"
	"strings"
	"time"

	"github.com/hashicorp/terraform-plugin-sdk/v2/diag"
	"github.com/hashicorp/terraform-plugin-sdk/v2/helper/schema"
	log "github.com/sirupsen/logrus"
	"repository.basistech.ru/BASIS/decort-golang-sdk/pkg/cloudbroker/disks"
	"repository.basistech.ru/BASIS/decort-golang-sdk/pkg/cloudbroker/image"
	"repository.basistech.ru/BASIS/decort-golang-sdk/pkg/cloudbroker/tasks"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/constants"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/controller"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/dc"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/status"
)

func resourceImageFromPlatformDiskCreate(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceImageFromPlatformDiskCreate: called for image %s", d.Get("name").(string))
	c := m.(*controller.ControllerCfg)

	if diags := checkParamsExistencePlatformDisk(ctx, d, c); diags != nil {
		return diags
	}

	req := disks.FromPlatformDiskRequest{
		DiskID:       uint64(d.Get("disk_id").(int)),
		Name:         d.Get("name").(string),
		BootType:     d.Get("boot_type").(string),
		ImageType:    d.Get("image_type").(string),
		Architecture: d.Get("architecture").(string),
		Bootable:     d.Get("bootable").(bool), // default is true
	}

	if username, ok := d.GetOk("username"); ok {
		req.Username = username.(string)
	}
	if password, ok := d.GetOk("password"); ok {
		req.Password = password.(string)
	}
	if accountId, ok := d.GetOk("account_id"); ok {
		req.AccountID = uint64(accountId.(int))
	}
	if poolName, ok := d.GetOk("pool_name"); ok {
		req.PoolName = poolName.(string)
	}
	if driversInterface, ok := d.GetOk("drivers"); ok {
		for _, d := range driversInterface.([]interface{}) {
			req.Drivers = append(req.Drivers, d.(string))
		}
	}
	if hotresize, ok := d.GetOk("hot_resize"); ok {
		req.HotResize = hotresize.(bool)
	}

	var imageId uint64
	var err error
	asyncMode := d.Get("async_mode").(bool)
	if !asyncMode {
		imageId, err = c.CloudBroker().Disks().FromPlatformDisk(ctx, req)
		if err != nil {
			d.SetId("")
			return diag.FromErr(err)
		}
	} else {
		taskId, err := c.CloudBroker().Disks().FromPlatformDiskAsync(ctx, req)
		if err != nil {
			d.SetId("")
			return diag.FromErr(err)
		}

		taskReq := tasks.GetRequest{
			AuditID: strings.Trim(taskId, `"`),
		}

		for {
			task, err := c.CloudBroker().Tasks().Get(ctx, taskReq)
			if err != nil {
				return diag.FromErr(err)
			}

			log.Debugf("resourceImageFromPlatformDiskCreate: instance creating - %s", task.Stage)

			if task.Completed {
				if task.Error != "" {
					return diag.FromErr(fmt.Errorf("cannot create image instance: %v", task.Error))
				}

				id, err := task.Result.ID()
				imageId = uint64(id)
				if err != nil {
					return diag.FromErr(err)
				}
				break
			}

			time.Sleep(time.Second * 20)
		}
	}

	d.SetId(strconv.FormatUint(imageId, 10))
	d.Set("image_id", imageId)

	var w dc.Warnings

	if _, ok := d.GetOk("accounts"); ok {
		if err := resourceImageSetAccess(ctx, d, m); err != nil {
			w.Add(err)
		}
	}

	return append(resourceImageFromPlatformDiskRead(ctx, d, m), w.Get()...)
}

func resourceImageFromPlatformDiskRead(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceImageFromPlatformDiskRead: called for %s id: %s", d.Get("name").(string), d.Id())

	img, err := utilityImageCheckPresence(ctx, d, m)
	if err != nil {
		d.SetId("")
		return diag.FromErr(err)
	}

	switch img.Status {
	case status.Modeled:
		return diag.Errorf("The image is in status: %s, please, contact support for more information", img.Status)
	case status.Destroyed, status.Purged:
		d.SetId("")
		return diag.Errorf("The resource cannot be updated because it has been destroyed")
	}

	flattenImage(d, img)

	return nil
}

func resourceImageFromPlatformDiskDelete(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceImageFromPlatformDiskDelete: called for %s, id: %s", d.Get("name").(string), d.Id())

	_, err := utilityImageCheckPresence(ctx, d, m)
	if err != nil {
		d.SetId("")
		return diag.FromErr(err)
	}

	c := m.(*controller.ControllerCfg)
	req := image.DeleteRequest{
		ImageID: uint64(d.Get("image_id").(int)),
	}

	_, err = c.CloudBroker().Image().Delete(ctx, req)
	if err != nil {
		return diag.FromErr(err)
	}

	d.SetId("")

	return nil
}

func resourceImageFromPlatformDiskUpdate(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("resourceImageFromPlatformDiskUpdate: called for %s, id: %s", d.Get("name").(string), d.Id())

	c := m.(*controller.ControllerCfg)
	if diags := checkParamsExistencePlatformDisk(ctx, d, c); diags != nil {
		return diags
	}

	// we do not allow change of disk_id, but allow resource update after import
	old, _ := d.GetChange("disk_id")
	if old.(int) != 0 && d.HasChange("disk_id") {
		return diag.Errorf("resourceImageFromPlatformDiskUpdate: can't update Image because disk_id is not allowed to be changed")
	}

	img, err := utilityImageCheckPresence(ctx, d, m)
	if err != nil {
		d.SetId("")
		return diag.FromErr(err)
	}

	switch img.Status {
	case status.Modeled:
		return diag.Errorf("The image is in status: %s, please, contact support for more information", img.Status)
	case status.Destroyed, status.Purged:
		d.SetId("")
		return diag.Errorf("The resource cannot be updated because it has been destroyed")
	}

	if d.HasChange("enabled") {
		err := resourceImageChangeEnabled(ctx, d, m)
		if err != nil {
			return diag.FromErr(err)
		}
	}

	if d.HasChange("shared_with") {
		err := resourceImageShare(ctx, d, m)
		if err != nil {
			return diag.FromErr(err)
		}
	}

	if d.HasChange("computeci_id") {
		err := resourceImageChangeComputeci(ctx, d, m)
		if err != nil {
			return diag.FromErr(err)
		}
	}

	if d.HasChange("enabled_stacks") {
		err := resourceImageUpdateNodes(ctx, d, m)
		if err != nil {
			return diag.FromErr(err)
		}
	}

	if d.HasChanges("name", "username", "password", "account_id", "bootable", "hot_resize", "network_interface_naming") {
		err := resourceImageEdit(ctx, d, m)
		if err != nil {
			return diag.FromErr(err)
		}
	}

	if d.HasChange("accounts") {
		err := resourceImageChangeAccess(ctx, d, m)
		if err != nil {
			return diag.FromErr(err)
		}
	}

	return resourceImageFromPlatformDiskRead(ctx, d, m)
}

func ResourceImageFromPlatformDisk() *schema.Resource {
	return &schema.Resource{
		SchemaVersion: 1,

		CreateContext: resourceImageFromPlatformDiskCreate,
		ReadContext:   resourceImageFromPlatformDiskRead,
		UpdateContext: resourceImageFromPlatformDiskUpdate,
		DeleteContext: resourceImageFromPlatformDiskDelete,

		Importer: &schema.ResourceImporter{
			StateContext: schema.ImportStatePassthroughContext,
		},

		Timeouts: &schema.ResourceTimeout{
			Create:  &constants.Timeout30m,
			Read:    &constants.Timeout900s,
			Update:  &constants.Timeout900s,
			Delete:  &constants.Timeout900s,
			Default: &constants.Timeout900s,
		},

		Schema: resourceImageFromPlatformDiskSchemaMake(),
	}
}
