/*
Copyright (c) 2019-2023 Digital Energy Cloud Solutions LLC. All Rights Reserved.
Authors:
Petr Krutov, <petr.krutov@digitalenergy.online>
Stanislav Solovev, <spsolovev@digitalenergy.online>
Kasim Baybikov, <kmbaybikov@basistech.ru>
Tim Tkachev, <tvtkachev@basistech.ru>

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

/*
Terraform DECORT provider - manage resources provided by DECORT (Digital Energy Cloud
Orchestration Technology) with Terraform by Hashicorp.

Source code: https://repository.basistech.ru/BASIS/terraform-provider-decort

Please see README.md to learn where to place source code so that it
builds seamlessly.

Documentation: https://repository.basistech.ru/BASIS/terraform-provider-decort/wiki
*/

package k8s

import (
	"context"
	"strconv"

	"github.com/hashicorp/terraform-plugin-sdk/v2/diag"
	"github.com/hashicorp/terraform-plugin-sdk/v2/helper/schema"
	log "github.com/sirupsen/logrus"

	"repository.basistech.ru/BASIS/decort-golang-sdk/pkg/cloudbroker/compute"
	"repository.basistech.ru/BASIS/decort-golang-sdk/pkg/cloudbroker/k8s"
	"repository.basistech.ru/BASIS/decort-golang-sdk/pkg/cloudbroker/lb"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/constants"
	"repository.basistech.ru/BASIS/terraform-provider-decort/internal/controller"
)

func dataSourceK8sRead(ctx context.Context, d *schema.ResourceData, m interface{}) diag.Diagnostics {
	log.Debugf("dataSourceK8sRead: called with k8s id %d", d.Get("k8s_id").(int))
	cluster, err := utilityK8sCheckPresence(ctx, d, m)
	if err != nil {
		d.SetId("")
		return diag.FromErr(err)
	}
	d.SetId(strconv.FormatUint(cluster.ID, 10))

	k8sList, err := utilityK8sListCheckPresence(ctx, d, m)
	if err != nil {
		d.SetId("")
		return diag.FromErr(err)
	}

	curK8s := k8s.ItemK8S{}
	for _, k8sCluster := range k8sList.Data {
		if k8sCluster.ID == cluster.ID {
			curK8s = k8sCluster
		}
	}
	if curK8s.ID == 0 {
		return diag.Errorf("Cluster with id %d not found in List clusters", cluster.ID)
	}

	d.Set("desc", curK8s.Description)
	d.Set("gid", curK8s.GID)
	d.Set("guid", curK8s.GUID)
	d.Set("milestones", curK8s.Milestones)
	d.Set("service_account", flattenServiceAccount(curK8s.ServiceAccount))
	d.Set("ssh_key", curK8s.SSHKey)
	d.Set("vins_id", curK8s.VINSID)

	masterComputeList := make([]compute.RecordCompute, 0, len(cluster.K8SGroups.Masters.DetailedInfo))
	workersComputeList := make([]compute.RecordCompute, 0, len(cluster.K8SGroups.Workers))
	for _, masterNode := range cluster.K8SGroups.Masters.DetailedInfo {
		compute, err := utilityComputeCheckPresence(ctx, d, m, masterNode.ID)
		if err != nil {
			return diag.FromErr(err)
		}
		masterComputeList = append(masterComputeList, *compute)
	}
	for _, worker := range cluster.K8SGroups.Workers {
		for _, info := range worker.DetailedInfo {
			compute, err := utilityComputeCheckPresence(ctx, d, m, info.ID)
			if err != nil {
				return diag.FromErr(err)
			}
			workersComputeList = append(workersComputeList, *compute)
		}
	}

	c := m.(*controller.ControllerCfg)

	getConfigReq := k8s.GetConfigRequest{K8SID: cluster.ID}

	kubeconfig, err := c.CloudBroker().K8S().GetConfig(ctx, getConfigReq)
	if err != nil {
		log.Warnf("could not get kubeconfig: %v", err)
	}

	d.Set("kubeconfig", kubeconfig)

	if cluster.LBID != 0 {
		getLbReq := lb.GetRequest{LBID: cluster.LBID}
		lb, err := c.CloudBroker().LB().Get(ctx, getLbReq)
		if err != nil {
			return diag.FromErr(err)
		}

		d.Set("extnet_id", lb.ExtNetID)
		d.Set("lb_ip", lb.PrimaryNode.FrontendIP)
	}

	flattenK8sData(d, cluster, masterComputeList, workersComputeList)
	return nil
}

func DataSourceK8s() *schema.Resource {
	return &schema.Resource{
		SchemaVersion: 1,

		ReadContext: dataSourceK8sRead,

		Timeouts: &schema.ResourceTimeout{
			Read:    &constants.Timeout30s,
			Default: &constants.Timeout60s,
		},

		Schema: dataSourceK8sSchemaMake(),
	}
}
